@@ -1,6 +1,5 @@
 cfg_io_driver! {
     pub(crate) mod bit;
-    pub(crate) mod slab;
 }
 
 #[cfg(feature = "rt")]
@@ -43,6 +42,10 @@ pub(crate) use wake_list::WakeList;
 ))]
 pub(crate) mod linked_list;
 
+cfg_rt! {
+    pub(crate) mod sharded_list;
+}
+
 #[cfg(any(feature = "rt", feature = "macros"))]
 pub(crate) mod rand;
 
@@ -71,3 +74,10 @@ cfg_rt_multi_thread! {
 pub(crate) mod trace;
 
 pub(crate) mod error;
+
+#[cfg(feature = "io-util")]
+pub(crate) mod memchr;
+
+pub(crate) mod markers;
+
+pub(crate) mod cacheline;
